@@ -121,6 +121,20 @@ memcache_delete(struct msg *r)
     return false;
 }
 
+/*
+ * Return true, if the memcache command is a touch command, otherwise
+ * return false
+ */
+static bool
+memcache_touch(struct msg *r)
+{
+    if (r->type == MSG_REQ_MC_TOUCH) {
+        return true;
+    }
+
+    return false;
+}
+
 void
 memcache_parse_req(struct msg *r)
 {
@@ -243,6 +257,14 @@ memcache_parse_req(struct msg *r)
 
                     break;
 
+                case 5:
+                    if (str5cmp(m, 't', 'o', 'u', 'c', 'h')) {
+                      r->type = MSG_REQ_MC_TOUCH;
+                      break;
+                    }
+
+                    break;
+
                 case 6:
                     if (str6cmp(m, 'a', 'p', 'p', 'e', 'n', 'd')) {
                         r->type = MSG_REQ_MC_APPEND;
@@ -282,6 +304,7 @@ memcache_parse_req(struct msg *r)
                 case MSG_REQ_MC_PREPEND:
                 case MSG_REQ_MC_INCR:
                 case MSG_REQ_MC_DECR:
+                case MSG_REQ_MC_TOUCH:
                     if (ch == CR) {
                         goto error;
                     }
@@ -320,13 +343,17 @@ memcache_parse_req(struct msg *r)
             }
             if (ch == ' ' || ch == CR) {
                 struct keypos *kpos;
-
-                if ((p - r->token) > MEMCACHE_MAX_KEY_LENGTH) {
+                int keylen = p - r->token;
+                if (keylen > MEMCACHE_MAX_KEY_LENGTH) {
                     log_error("parsed bad req %"PRIu64" of type %d with key "
                               "prefix '%.*s...' and length %d that exceeds "
                               "maximum key length", r->id, r->type, 16,
                               r->token, p - r->token);
                     goto error;
+                } else if (keylen == 0) {
+                    log_error("parsed bad req %"PRIu64" of type %d with an "
+                              "empty key", r->id, r->type);
+                    goto error;
                 }
 
                 kpos = array_push(r->keys);
@@ -342,7 +369,7 @@ memcache_parse_req(struct msg *r)
                 /* get next state */
                 if (memcache_storage(r)) {
                     state = SW_SPACES_BEFORE_FLAGS;
-                } else if (memcache_arithmetic(r)) {
+                } else if (memcache_arithmetic(r) || memcache_touch(r) ) {
                     state = SW_SPACES_BEFORE_NUM;
                 } else if (memcache_delete(r)) {
                     state = SW_RUNTO_CRLF;
@@ -531,7 +558,7 @@ memcache_parse_req(struct msg *r)
 
         case SW_SPACES_BEFORE_NUM:
             if (ch != ' ') {
-                if (!isdigit(ch)) {
+                if (!(isdigit(ch) || ch == '-')) {
                     goto error;
                 }
                 /* num_start <- p; num <- ch - '0'  */
@@ -562,7 +589,7 @@ memcache_parse_req(struct msg *r)
                 break;
 
             case 'n':
-                if (memcache_storage(r) || memcache_arithmetic(r) || memcache_delete(r)) {
+                if (memcache_storage(r) || memcache_arithmetic(r) || memcache_delete(r) || memcache_touch(r)) {
                     /* noreply_start <- p */
                     r->token = p;
                     state = SW_NOREPLY;
@@ -593,7 +620,7 @@ memcache_parse_req(struct msg *r)
             case CR:
                 m = r->token;
                 if (((p - m) == 7) && str7cmp(m, 'n', 'o', 'r', 'e', 'p', 'l', 'y')) {
-                    ASSERT(memcache_storage(r) || memcache_arithmetic(r) || memcache_delete(r));
+                    ASSERT(memcache_storage(r) || memcache_arithmetic(r) || memcache_delete(r) || memcache_touch(r));
                     r->token = NULL;
                     /* noreply_end <- p - 1 */
                     r->noreply = 1;
@@ -854,6 +881,11 @@ memcache_parse_rsp(struct msg *r)
                         break;
                     }
 
+                    if (str7cmp(m, 'T', 'O', 'U', 'C', 'H', 'E', 'D')) {
+                        r->type = MSG_RSP_MC_TOUCHED;
+                        break;
+                    }
+
                     break;
 
                 case 9:
@@ -895,6 +927,7 @@ memcache_parse_rsp(struct msg *r)
                 case MSG_RSP_MC_EXISTS:
                 case MSG_RSP_MC_NOT_FOUND:
                 case MSG_RSP_MC_DELETED:
+                case MSG_RSP_MC_TOUCHED:
                     state = SW_CRLF;
                     break;
 
@@ -1167,6 +1200,12 @@ error:
                 r->state);
 }
 
+bool
+memcache_failure(struct msg *r)
+{
+    return false;
+}
+
 static rstatus_t
 memcache_append_key(struct msg *r, uint8_t *key, uint32_t keylen)
 {
@@ -1377,9 +1416,8 @@ memcache_pre_coalesce(struct msg *r)
 }
 
 /*
- * copy one response from src to dst
- * return bytes copied
- * */
+ * Copy one response from src to dst and return bytes copied
+ */
 static rstatus_t
 memcache_copy_bulk(struct msg *dst, struct msg *src)
 {
@@ -1403,9 +1441,10 @@ memcache_copy_bulk(struct msg *dst, struct msg *src)
     }
     p = mbuf->pos;
 
-    /* get : VALUE key 0 len\r\nval\r\n */
-    /* gets: VALUE key 0 len cas\r\nval\r\n */
-
+    /*
+     * get : VALUE key 0 len\r\nval\r\n
+     * gets: VALUE key 0 len cas\r\nval\r\n
+     */
     ASSERT(*p == 'V');
     for (i = 0; i < 3; i++) {                 /*  eat 'VALUE key 0 '  */
         for (; *p != ' ';) {
@@ -1495,3 +1534,28 @@ memcache_post_coalesce(struct msg *request)
         return;
     }
 }
+
+void
+memcache_post_connect(struct context *ctx, struct conn *conn, struct server *server)
+{
+}
+
+void
+memcache_swallow_msg(struct conn *conn, struct msg *pmsg, struct msg *msg)
+{
+}
+
+rstatus_t
+memcache_add_auth(struct context *ctx, struct conn *c_conn, struct conn *s_conn)
+{
+    NOT_REACHED();
+    return NC_OK;
+}
+
+rstatus_t
+memcache_reply(struct msg *r)
+{
+    NOT_REACHED();
+    return NC_OK;
+}
+
